from dataclasses import dataclass
from enum import Enum

class RelType(Enum):
    COMPOSITION = '*'
    INHERITANCE = '<|'

@dataclass
class UmlRelation(object):
    source_fqdn: str
    target_fqdn: str
    type: RelType



from py2puml.py2puml import py2puml

# outputs the PlantUML content in the terminal
print(''.join(
    py2puml('py2puml/domain', 'py2puml.domain')
))

# writes the PlantUML content in a file
with open('py2puml/domain.puml', 'w') as puml_file:
    puml_file.writelines(py2puml('py2puml/domain', 'py2puml.domain'))



from typing import List
from dataclasses import dataclass

from py2puml.domain.umlitem import UmlItem

@dataclass
class Member(object):
    name: str
    value: str

@dataclass
class UmlEnum(UmlItem):
    members: List[Member]



from typing import List
from dataclasses import dataclass

from py2puml.domain.umlitem import UmlItem

@dataclass
class UmlAttribute(object):
    name: str
    type: str

@dataclass
class UmlClass(UmlItem):
    attributes: List[UmlAttribute]



from typing import Dict, List, Iterable
from types import ModuleType
from pkgutil import walk_packages
from importlib import import_module

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.parser import parse_item_module
from py2puml.exportpuml import to_puml_content


def py2puml(domain_path: str, domain_module: str) -> Iterable[str]:
    domain_items_by_fqdn: Dict[str, UmlItem] = {}
    domain_relations: List[UmlRelation] = []
    for _, name, is_pkg in walk_packages([domain_path]):
        if not is_pkg:
            domain_item_module: ModuleType = import_module(
                f'{domain_module}.{name}'
            )
            parse_item_module(
                domain_item_module,
                domain_module,
                domain_items_by_fqdn,
                domain_relations
            )
    return to_puml_content(domain_items_by_fqdn.values(), domain_relations)



from typing import Iterable, Type, List, Dict
from types import ModuleType
from enum import Enum
from inspect import getmembers
from re import compile

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlclass import UmlClass, UmlAttribute
from py2puml.domain.umlenum import UmlEnum, Member
from py2puml.domain.umlrelation import UmlRelation, RelType
from py2puml.utils import inspect_type

CONCRETE_TYPE_PATTERN = compile("^<(?:class|enum) '([\\.|\\w]+)'>$")

def filter_domain_definitions(module: ModuleType, root_module_name: str) -> Iterable[Type]:
    for definition_key in dir(module):
        definition_type = getattr(module, definition_key)
        definition_members = getmembers(definition_type)
        definition_module_member = next(
            (member for member in definition_members if member[0] == '__module__' and member[1].startswith(root_module_name)),
            None
        )
        if definition_module_member is not None:
            yield definition_type

def get_type_name(type: Type, root_module_name: str):
    if type.__module__.startswith(root_module_name):
        return type.__name__
    else:
        return f'{type.__module__}.{type.__name__}'

def parse_enum_type(
    enum_type: Type,
    enum_type_fqdn: str,
    domain_items_by_fqdn: Dict[str, UmlItem]
):
    domain_items_by_fqdn[enum_type_fqdn] = UmlEnum(
        name=enum_type.__name__,
        fqdn=enum_type_fqdn,
        members=[
            Member(name=enum_member.name, value=enum_member.value)
            for enum_member in enum_type.__members__.values()
        ]
    )

def parse_namedtupled_class(
    namedtupled_type: Type,
    namedtupled_type_fqdn: str,
    domain_items_by_fqdn: Dict[str, UmlItem]
):
    domain_items_by_fqdn[namedtupled_type_fqdn] = UmlClass(
        name=namedtupled_type.__name__,
        fqdn=namedtupled_type_fqdn,
        attributes=[
            UmlAttribute(tuple_field, 'any')
            for tuple_field in namedtupled_type._fields
        ]
    )


def handle_inheritance_relation(
    class_type: Type,
    class_fqdn: str,
    root_module_name: str,
    domain_relations: List[UmlRelation]
):
    for base_type in getattr(class_type, '__bases__', ()):
        base_type_fqdn = f'{base_type.__module__}.{base_type.__name__}'
        if base_type_fqdn.startswith(root_module_name):
            domain_relations.append(
                UmlRelation(base_type_fqdn, class_fqdn, RelType.INHERITANCE)
            )

def parse_class_type(
    class_type: Type,
    class_type_fqdn: str,
    root_module_name: str,
    domain_items_by_fqdn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    definition_attrs = []
    uml_class = UmlClass(
        name=class_type.__name__,
        fqdn=class_type_fqdn,
        attributes=definition_attrs
    )
    domain_items_by_fqdn[class_type_fqdn] = uml_class
    # inspect_type(class_type)
    type_annotations = getattr(class_type, '__annotations__', None)
    if type_annotations is not None:
        for attr_name, attr_class in type_annotations.items():
            attr_raw_type = str(attr_class)
            concrete_type_match = CONCRETE_TYPE_PATTERN.search(attr_raw_type)
            if concrete_type_match:
                concrete_type = concrete_type_match.group(1)
                if attr_class.__module__.startswith(root_module_name):
                    attr_type = attr_class.__name__
                    domain_relations.append(
                        UmlRelation(uml_class.fqdn, f'{attr_class.__module__}.{attr_class.__name__}', RelType.COMPOSITION)
                    )
                else:
                    attr_type = concrete_type
            else:
                composition_rel = getattr(attr_class, '_name', None)
                component_classes = getattr(attr_class, '__args__', None)
                if composition_rel and component_classes:
                    component_names = [
                        get_type_name(component_class, root_module_name)
                        for component_class in component_classes
                        # filters out forward refs
                        if getattr(component_class, '__name__', None) is not None
                    ]
                    domain_relations.extend([
                        UmlRelation(uml_class.fqdn, f'{component_class.__module__}.{component_class.__name__}', RelType.COMPOSITION)
                        for component_class in component_classes
                        if component_class.__module__.startswith(root_module_name)
                    ])
                    attr_type = f"{composition_rel}[{', '.join(component_names)}]"
                else:
                    attr_type = attr_raw_type
            uml_attr = UmlAttribute(attr_name, attr_type)
            definition_attrs.append(uml_attr)

    handle_inheritance_relation(class_type, class_type_fqdn, root_module_name, domain_relations)


def parse_type(
    definition_type: Type,
    root_module_name: str,
    domain_items_by_fqdn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    definition_type_fqdn = f'{definition_type.__module__}.{definition_type.__name__}'
    if definition_type_fqdn not in domain_items_by_fqdn:
        if issubclass(definition_type, Enum):
            parse_enum_type(definition_type, definition_type_fqdn, domain_items_by_fqdn)
        elif getattr(definition_type, '_fields', None) is not None:
            parse_namedtupled_class(definition_type, definition_type_fqdn, domain_items_by_fqdn)
        else:
            parse_class_type(definition_type, definition_type_fqdn, root_module_name, domain_items_by_fqdn, domain_relations)


def parse_item_module(
    domain_item_module: ModuleType,
    root_module_name: str,
    domain_items_by_fqdn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    # processes only the classes imported and defined within the given root module
    for definition_type in filter_domain_definitions(domain_item_module, root_module_name):
        parse_type(definition_type, root_module_name, domain_items_by_fqdn, domain_relations)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py2puml', 'py2puml.domain']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'py2puml',
    'version': '0.3.0',
    'description': 'Generate Plantuml diagrams to document your python code ',
    'long_description': '<div align="center">\n  <a href="https://www.python.org/psf-landing/" target="_blank">\n    <img width="350px" alt="Python logo"\n      src="https://www.python.org/static/community_logos/python-logo-generic.svg" />\n  </a>\n  <a href="http://plantuml.com/" target="_blank">\n    <img width="116px" height="112px" alt="PlantUML logo" src="http://s.plantuml.com/logoc.png" style="margin-bottom: 40px" vspace="40px" />\n  </a>\n  <h1>Python to PlantUML</h1>\n</div>\n\nGenerate Plantuml diagrams to document your python code\n\n# How it works\n\n## Features\n\nFrom a given path corresponding to a folder containing python code, `py2puml` loads each file as a module and generate a class diagram with the [PlantUML](https://plantuml.com/en/class-diagram) using:\n\n* inspection to detect the classes to document (see the [inspect](https://docs.python.org/3/library/inspect.html) module)\n* annotations (the python type hinting syntax) to detect the attributes and their types (see the [typing](https://docs.python.org/3/library/typing.html) module)\n* fields for classes derived from namedtuples\n* composition and inheritance relationships are drawn only between the domain classes (this is designed on purpose, for documentation sake)\n\n## Current limitations\n\n* type hinting is optional when writing Python code and discarded when it is executed, as mentionned in the [typing official documentation](https://docs.python.org/3/library/typing.html). The quality of the diagram output by `py2puml` depends on the reliability with which the type annotations were written\n\n> The Python runtime does not enforce function and variable type annotations. They can be used by third party tools such as type checkers, IDEs, linters, etc.\n\n* complex type hints with more than one level of genericity are not properly handled for the moment: `List[MyClass]` or `Dict[str, MyClass]` are handled properly, `Dict[str, List[MyClass]]` is not. If your domain classes (also called business objects or DTOs) have attributes with complex type hints, it may be a code smell indicating that you should write a class which would better represent the business logic. But I may improve this part of the library as well 😀\n\n* `py2puml` does not inspect sub-folders recursively, but it is planned\n\n* `py2puml` outputs diagrams in PlantUML syntax, which can be saved in text files along your python code and versioned with them. To generate image files, use the PlantUML runtime or a docker image (see [think/plantuml](https://hub.docker.com/r/think/plantuml))\n\n* `py2puml` uses features of python 3 (generators for example) and thus won\'t work with python 2 runtimes. It relies on native python modules and uses no 3rd-party library, except [pytest](https://docs.pytest.org/en/latest/) as a development dependency for running the unit-tests\n\nIf you like tools around PlantUML, you may also be interested in this [lucsorel/plantuml-file-loader](https://github.com/lucsorel/plantuml-file-loader) project: A webpack loader which converts PlantUML files into images during the webpack processing (useful to [include PlantUML diagrams in your slides](https://github.com/lucsorel/markdown-image-loader/blob/master/README.md#web-based-slideshows) with RevealJS or RemarkJS).\n\n# Install\n\nInstall from the github repository:\n\n* with `pip`:\n\n```sh\npip install py2puml\n```\n\n* with [poetry](https://pipenv.readthedocs.io/en/latest/):\n\n```sh\npoetry add py2puml\n```\n\n* with [pipenv](https://pipenv.readthedocs.io/en/latest/):\n\n```sh\npipenv install py2puml\n```\n\n# Usage\n\nFor example, to create the diagram of the classes used by `py2puml`:\n\n* import the py2puml function in your script (see [py2puml/example.py](py2puml/example.py)):\n\n```python\nfrom py2puml.py2puml import py2puml\n\n# outputs the PlantUML content in the terminal\nprint(\'\'.join(py2puml(\'py2puml/domain\', \'py2puml.domain\')))\n\n# writes the PlantUML content in a file\nwith open(\'py2puml/domain.puml\', \'w\') as puml_file:\n    puml_file.writelines(py2puml(\'py2puml/domain\', \'py2puml.domain\'))\n```\n\n* running it (`python3 -m py2puml.example`) will output the PlantUML diagram in the terminal and write it in a file\n\n```plantuml\n@startuml\nclass py2puml.domain.umlclass.UmlAttribute {\n  name: str\n  type: str\n}\nclass py2puml.domain.umlclass.UmlClass {\n  attributes: List[UmlAttribute]\n}\nclass py2puml.domain.umlitem.UmlItem {\n  name: str\n  fqdn: str\n}\nclass py2puml.domain.umlenum.Member {\n  name: str\n  value: str\n}\nclass py2puml.domain.umlenum.UmlEnum {\n  members: List[Member]\n}\nenum py2puml.domain.umlrelation.RelType {\n  COMPOSITION: *\n  INHERITANCE: <|\n}\nclass py2puml.domain.umlrelation.UmlRelation {\n  source_fqdn: str\n  target_fqdn: str\n  type: RelType\n}\npy2puml.domain.umlclass.UmlClass *-- py2puml.domain.umlclass.UmlAttribute\npy2puml.domain.umlitem.UmlItem <|-- py2puml.domain.umlclass.UmlClass\npy2puml.domain.umlenum.UmlEnum *-- py2puml.domain.umlenum.Member\npy2puml.domain.umlitem.UmlItem <|-- py2puml.domain.umlenum.UmlEnum\npy2puml.domain.umlrelation.UmlRelation *-- py2puml.domain.umlrelation.RelType\n@enduml\n```\n\nWhich renders like this:\n\n![](https://www.plantuml.com/plantuml/png/ZP91IyGm48Nl-HKvBsmF7iiUTbaA1jnMQZs9I7OxIY19Qp8H5jV_xZIse5GsFULrQBvvCozRZz9XC9gTjFIUz-URdhwojZDIsOnah6UFHkyGdJe61Fx9EBVIGCuzEj9uxaVzbSRi1n4HSWBwdDyfZq-_cpnVOIa4Cw04dJCph--jJPa16qns07C4Dxl_8NM0HG1oKD0P2IR2fa5-qCC8mu__t7UW9QhEPZNeXhON6VlgS5yzY4PKPSvNL13bRL6BPbVkYvnlBdC_SnvvgaSTcRuBxWGlSIbJMjAz0SRItm17BzGc6TzglLxqL5WYlCs5GAbkBB5_CdCzuoKk4Y6pPJkFNj9niotObkhi6m00)\n\n# Tests\n\n```sh\npython3 -m pytest -v\n```\n\n# Changelog\n\n* `0.3.0`: handle classes derived from namedtuples (attribute types are `any`)\n* `0.2.0`: handle inheritance relationships and enums. Unit tested\n* `0.1.3`: first release, handle all module of a folder and compositions of domain classes\n\n# Licence\n\nUnless stated otherwise all works are licensed under the [MIT license](http://spdx.org/licenses/MIT.html), a copy of which is included [here](LICENSE).\n\n# Contributions\n\n* [Luc Sorel-Giffo](https://github.com/lucsorel)\n\nPull-requests are welcome and will be processed on a best-effort basis.\n\n\n# Alternatives\n\nIf `py2puml` does not meet your needs (suggestions and pull-requests are **welcome**), you can have a look at these projects which follow other approaches (AST, linting, modeling):\n\n* [cb109/pyplantuml](https://github.com/cb109/pyplantuml)\n* [deadbok/py-puml-tools](https://github.com/deadbok/py-puml-tools)\n* [caballero/genUML](https://github.com/jose-caballero/genUML)\n',
    'author': 'Luc Sorel-Giffo',
    'author_email': None,
    'maintainer': 'Luc Sorel-Giffo',
    'maintainer_email': None,
    'url': 'https://github.com/lucsorel/py2puml',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from os.path import dirname, realpath, join
from re import compile, Match

# exports the version read from the pyproject.toml file
try:
    PARENT_DIR = dirname(dirname(realpath(__file__)))
    VERSION_PATTERN = compile('^version = "([^"]+)"$')
    with open(join(PARENT_DIR, 'pyproject.toml')) as pyproject:
        version_match: Match = next((
            match for match in (VERSION_PATTERN.search(line) for line in pyproject.readlines())
            if match is not None
        ), None)

        __version__ = version_match.group(1)
except:
    __version__ = None




from typing import List, Iterable

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlclass import UmlClass
from py2puml.domain.umlenum import UmlEnum
from py2puml.domain.umlrelation import UmlRelation

PUML_FILE_START = '@startuml\n'
PUML_FILE_END = '@enduml\n'
PUML_ITEM_START_TPL = '{item_type} {item_fqdn} {{\n'
PUML_ATTR_TPL = '  {attr_name}: {attr_type}\n'
PUML_ITEM_END = '}\n'
PUML_COMPOSITION_TPL = '{source_fqdn} {rel_type}-- {target_fqdn}\n'


def to_puml_content(uml_items: List[UmlItem], uml_relations: List[UmlRelation]) -> Iterable[str]:
    yield PUML_FILE_START

    # exports the domain classes and enums
    for uml_item in uml_items:
        if isinstance(uml_item, UmlEnum):
            uml_enum: UmlEnum = uml_item
            yield PUML_ITEM_START_TPL.format(item_type='enum', item_fqdn=uml_enum.fqdn)
            for member in uml_enum.members:
                yield PUML_ATTR_TPL.format(attr_name=member.name, attr_type=member.value)
            yield PUML_ITEM_END
        elif isinstance(uml_item, UmlClass):
            uml_class: UmlClass = uml_item
            yield PUML_ITEM_START_TPL.format(item_type='class', item_fqdn=uml_class.fqdn)
            for uml_attr in uml_class.attributes:
                yield PUML_ATTR_TPL.format(attr_name=uml_attr.name, attr_type=uml_attr.type)
            yield PUML_ITEM_END
        else:
            raise TypeError(f'cannot process uml_item of type {uml_item.__class__}')

    # exports the domain relationships between classes and enums
    for uml_relation in uml_relations:
        yield PUML_COMPOSITION_TPL.format(
            source_fqdn=uml_relation.source_fqdn,
            rel_type=uml_relation.type.value,
            target_fqdn=uml_relation.target_fqdn
        )

    yield PUML_FILE_END
